searchNodes=[{"doc":"Utilities to work with Elixir source code. NOTICE: This library is under heavy development. Expect frequent breaking changes until the first stable v1.0 release is out. Installation Add :sourceror as a dependency to your project's mix.exs : defp deps do [ { :sourceror , &quot;~&gt; 0.14&quot; } ] end A note on compatibility Sourceror is compatible with Elixir versions down to 1.10 and OTP 21. For Elixir versions prior to 1.13 it uses a vendored version of the Elixir parser and formatter modules. This means that for Elixir versions prior to 1.12 it will successfully parse the new syntax for stepped ranges instead of raising a SyntaxError , but everything else should work as expected. Goals of the library Be as close as possible to the standard Elixir AST. Make working with comments as simple as possible. No dev/prod dependencies, to simplify integration with other tools. Sourceror's AST Having the AST and comments as separate entities allows Elixir to expose the code formatting utilities without making any changes to it's AST, but also delegates the task of figuring out what's the most appropriate way to work with them to us. Sourceror's take is to use the node metadata to store the comments. This allows us to work with an AST that is as close to regular elixir AST as possible. It also allows you to move nodes around without worrying about leaving a comment behind and ending up with misplaced comments. Two metadata fields are added to the regular Elixir AST: :leading_comments - holds the comments directly above the node or are in the same line as it. For example: test &quot;parses leading comments&quot; do quoted = &quot;&quot;&quot; # Comment for :a :a # Also a comment for :a &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) assert { :__block__ , meta , [ :a ] } = quoted assert meta [ :leading_comments ] == [ %{ line : 1 , column : 1 , previous_eol_count : 1 , next_eol_count : 1 , text : &quot;# Comment for :a&quot; } , %{ line : 2 , column : 4 , previous_eol_count : 0 , next_eol_count : 1 , text : &quot;# Also a comment for :a&quot; } , ] end :trailing_comments - holds the comments that are inside of the node, but aren't leading any children, for example: test &quot;parses trailing comments&quot; do quoted = &quot;&quot;&quot; def foo() do :ok # A trailing comment end # Not a trailing comment for :foo &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) assert { :__block__ , block_meta , [ { :def , meta , _ } ] } = quoted assert [ %{ line : 3 , text : &quot;# A trailing comment&quot; } ] = meta [ :trailing_comments ] assert [ %{ line : 4 , text : &quot;# Not a trailing comment for :foo&quot; } ] = block_meta [ :trailing_comments ] end Note that Sourceror considers leading comments to the ones that are found in the same line as a node, and trailing comments to the ones that are found before the ending line of a node, based on the end , closing or end_of_expression line. This also makes the Sourceror AST consistent with the way the Elixir formatter works, making it easier to reason about how a given AST would be formatted. Traversing the AST Elixir provides the Macro.prewalk , Macro.postwalk and Macro.traverse functions to traverse the AST. You can use the same functions to traverse the Sourceror AST as well, since it has the same shape as the standard Elixir AST. Sourceror also provides the Sourceror.prewalk , Sourceror.postwalk and Sourceror.traverse variants. At the time of writing they are mostly wrappers around the standard Elixir functions for AST traversal, but they may be enhanced in the future if more AST formats are introduced. In addition to these, Sourceror also provides a Zipper implementation for the Elixir AST. You can learn more about it in the Zippers notebook . Patching the source code You can use Sourceror to manipulate the AST and turn it back into human readable Elixir code, this is commonly known as writing a &quot;codemod&quot;. For example, you can write a codemod to replace calls to String.to_atom to String.to_existing_atom : test &quot;updates the source code&quot; do source = &quot;&quot;&quot; String.to_atom(foo) \\ &quot;&quot;&quot; new_source = source |&gt; Sourceror . parse_string! ( ) |&gt; Macro . postwalk ( fn { { :. , dot_meta , [ { :__aliases__ , alias_meta , [ :String ] } , :to_atom ] } , call_meta , args } -&gt; { { :. , dot_meta , [ { :__aliases__ , alias_meta , [ :String ] } , :to_existing_atom ] } , call_meta , args } quoted -&gt; quoted end ) |&gt; Sourceror . to_string ( ) assert new_source == &quot;&quot;&quot; String.to_existing_atom(foo) \\ &quot;&quot;&quot; end However, this will affect the whole source code, as we are working on the full source AST. Sourceror relies on the Elixir formatter to produce human readable code, so the original code formatting will be lost by using Sourceror.to_string . If your code is already using the Elixir formatter then this won't be an issue, but it will be an undesirable effect if you're not using it. An alternative to this is to use Patches instead. A patch is a data structure that specifies the text range that should be replaced, and either a replacement string or a function that takes the text in that range and produces a string replacement. Using patches, we could do the same as above, but produce a patch instead of modifying the AST. As a result, only the parts that need to be changed will be affected, and the rest of the code keeps the original formatting: test &quot;patches the source code&quot; do source = &quot;&quot;&quot; case foo do nil -&gt; :bar _ -&gt; String.to_atom(foo) end \\ &quot;&quot;&quot; { _quoted , patches } = source |&gt; Sourceror . parse_string! ( ) |&gt; Macro . postwalk ( [ ] , fn { { :. , dot_meta , [ { :__aliases__ , alias_meta , [ :String ] } , :to_atom ] } , call_meta , args } = quoted , patches -&gt; range = Sourceror . get_range ( quoted ) replacement = { { :. , dot_meta , [ { :__aliases__ , alias_meta , [ :String ] } , :to_existing_atom ] } , call_meta , args } |&gt; Sourceror . to_string ( ) patch = %{ range : range , change : replacement } { quoted , [ patch | patches ] } quoted , patches -&gt; { quoted , patches } end ) assert Sourceror . patch_string ( source , patches ) == &quot;&quot;&quot; case foo do nil -&gt; :bar _ -&gt; String.to_existing_atom(foo) end \\ &quot;&quot;&quot; end You have to keep in mind that: If you patch a node that has inner code, like replacing a full case , then the contents of the node will be reformatted as well. At the moment, Sourceror won't check for conflicts in the patches ranges, so care needs to be taken to not produce conflicting patches. You may need to do a number of parse -&gt; patch -&gt; reparse if you find yourself generating conflicting patches. Some of the most common patching operations are available in the Sourceror.Patch module Background There have been several attempts at source code manipulation in the Elixir community. Thanks to its metaprogramming features, Elixir provides builtin tools that let us get the AST of any Elixir code, but when it comes to turning the AST back to code as text, we had limited options. Macro.to_string/2 is a thing, but the produced code is generally ugly, mostly because of the extra parenthesis or because it turns string interpolations into calls to erlang modules, to name some examples. This meant that, even if we could use Macro.to_string/2 to get a string and then give that to the Elixir formatter Code.format_string!/2 , the output would still be suboptimal, as the formatter is not designed to change the semantics of the code, only to pretty print it. For example, call to erlang modules would be kept as is instead of being turned back to interpolations. We also had the additional problem of comments being discarded by the tokenizer, and literals not having information like line numbers or delimiter characters. This makes the regular AST too lossy to be useful if what we want is to manipulate the source code, because we need as much information as possible to be able to stay as close to the source as possible. There have been several proposal in the past to bring all this information to the Elixir AST, but they all meant a change that would either break macros due to the addition of new types of AST nodes, or making a compromise in core Elixir itself by storing comments in the nods metadata. This discussion in the Elixir mailing list highlights the various issues faced when deciding if and how the comments would be preserved. Arjan Scherpenisse also did a talk where he discusses about the problems of using the standard Elixir AST to build refactoring tools. Despite of all these issues, the Elixir formatter is still capable of manipulating the source code to pretty print it. Under the hood it does some neat tricks to have all this information available: on one hand, it tells the tokenizer to extract the comments from the source code and keep it at hand(not in the AST itself, but as a separate data structure), and on the other hand it tells the parser to wrap literals in block nodes so metadata can be preserved. Once it has all it needs, it can start converting the AST and comments into an algebra document, and ultimately convert that to a string. This functionality was private, and if we wanted to do it ourselves we would have to replicate or vendor the Elixir formatter with all its more than 2000 lines of code. This approach was explored by Wojtek Mach in wojtekmach/fix , but it involved vendoring the elixir Formatter code, was tightly coupled to the formatting process, and any change in Elixir would break the code. Since Elixir 1.13 this functionality from the formatter was finally exposed via the Code.string_to_quoted_with_comments/2 and Code.quoted_to_algebra/2 functions. The former gives us access to the list of comments in a shape the Elixir formatter is able to use, and the latter lets us turn any arbitrary Elixir AST into an algebra document. If we also give it the list of comments, it will merge them together, allowing us to format AST and preserve the comments. Now all we need to care about is of manipulating the AST, and let the formatter do the rest.","ref":"Sourceror.html","title":"Sourceror","type":"module"},{"doc":"Appends comments to the leading or trailing comments of a node.","ref":"Sourceror.html#append_comments/3","title":"Sourceror.append_comments/3","type":"function"},{"doc":"Compares two positions. Returns :gt if the first position comes after the second one, and :lt for vice versa. If the two positions are equal, :eq is returned. nil values for lines or columns are coalesced to 0 for integer comparisons.","ref":"Sourceror.html#compare_positions/2","title":"Sourceror.compare_positions/2","type":"function"},{"doc":"Shifts the line numbers of the node or metadata by the given line_correction . This function will update the :line , :closing , :do , :end and :end_of_expression line numbers of the node metadata if such fields are present.","ref":"Sourceror.html#correct_lines/3","title":"Sourceror.correct_lines/3","type":"function"},{"doc":"Returns the arguments of the node. iex&gt; Sourceror . get_args ( { :foo , [ ] , [ { :__block__ , [ ] , [ :ok ] } ] } ) [ { :__block__ , [ ] , [ :ok ] } ]","ref":"Sourceror.html#get_args/1","title":"Sourceror.get_args/1","type":"function"},{"doc":"Returns the column of a node. If none is found, the default value is returned(defaults to 1). A default of nil may also be provided if the column number is meant to be coalesced with a value that is not known upfront. iex&gt; Sourceror . get_column ( { :foo , [ column : 5 ] , [ ] } ) 5 iex&gt; Sourceror . get_column ( { :foo , [ ] , [ ] } , 3 ) 3","ref":"Sourceror.html#get_column/2","title":"Sourceror.get_column/2","type":"function"},{"doc":"Returns the line where the given node ends. It recursively checks for end , closing and end_of_expression line numbers. If none is found, the default value is returned(defaults to 1). iex&gt; Sourceror . get_end_line ( { :foo , [ end : [ line : 4 ] ] , [ ] } ) 4 iex&gt; Sourceror . get_end_line ( { :foo , [ closing : [ line : 2 ] ] , [ ] } ) 2 iex&gt; Sourceror . get_end_line ( { :foo , [ end_of_expression : [ line : 5 ] ] , [ ] } ) 5 iex&gt; Sourceror . get_end_line ( { :foo , [ closing : [ line : 2 ] , end : [ line : 4 ] ] , [ ] } ) 4 iex&gt; &quot;&quot;&quot; ...&gt; alias Foo.{ ...&gt; Bar ...&gt; } ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) |&gt; Sourceror . get_end_line ( ) 3","ref":"Sourceror.html#get_end_line/2","title":"Sourceror.get_end_line/2","type":"function"},{"doc":"Returns the end position of the quoted expression. It recursively checks for end , closing and end_of_expression positions. If none is found, the default value is returned(defaults to [line: 1, column: 1] ). iex&gt; quoted = ~S&quot;&quot;&quot; ...&gt; A.{ ...&gt; B ...&gt; } ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) iex&gt; Sourceror . get_end_position ( quoted ) [ line : 3 , column : 1 ] iex&gt; quoted = ~S&quot;&quot;&quot; ...&gt; foo do ...&gt; :ok ...&gt; end ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) iex&gt; Sourceror . get_end_position ( quoted ) [ line : 3 , column : 1 ] iex&gt; quoted = ~S&quot;&quot;&quot; ...&gt; foo( ...&gt; :a, ...&gt; :b ...&gt; ) ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) iex&gt; Sourceror . get_end_position ( quoted ) [ line : 4 , column : 4 ]","ref":"Sourceror.html#get_end_position/2","title":"Sourceror.get_end_position/2","type":"function"},{"doc":"Returns the line of a node. If none is found, the default value is returned(defaults to 1). A default of nil may also be provided if the line number is meant to be coalesced with a value that is not known upfront. iex&gt; Sourceror . get_line ( { :foo , [ line : 5 ] , [ ] } ) 5 iex&gt; Sourceror . get_line ( { :foo , [ ] , [ ] } , 3 ) 3","ref":"Sourceror.html#get_line/2","title":"Sourceror.get_line/2","type":"function"},{"doc":"Returns the metadata of the given node. iex&gt; Sourceror . get_meta ( { :foo , [ line : 5 ] , [ ] } ) [ line : 5 ]","ref":"Sourceror.html#get_meta/1","title":"Sourceror.get_meta/1","type":"function"},{"doc":"Gets the range used by the given quoted expression in the source code. The quoted expression must have at least line and column metadata, otherwise it is not possible to calculate an accurate range, or to calculate it at all. Additionally, certain syntax constructs desugar into ASTs without a meaningful range. In these cases, get_range/1 returns nil . This function is most useful when used after Sourceror.parse_string/1 , before any kind of modification to the AST. The range is a map with :start and :end positions. iex&gt; quoted = ~S&quot;&quot;&quot; ...&gt; def foo do ...&gt; :ok ...&gt; end ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) iex&gt; Sourceror . get_range ( quoted ) %{ start : [ line : 1 , column : 1 ] , end : [ line : 3 , column : 4 ] } iex&gt; quoted = ~S&quot;&quot;&quot; ...&gt; Foo.{ ...&gt; Bar ...&gt; } ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) iex&gt; Sourceror . get_range ( quoted ) %{ start : [ line : 1 , column : 1 ] , end : [ line : 3 , column : 2 ] } Options - ` :include_comments ` - When ` true ` , it includes the comments into the range . Defaults to ` false ` . iex&gt; ~S&quot;&quot;&quot; ...&gt; # Foo ...&gt; :baz # Bar ...&gt; &quot;&quot;&quot; ...&gt; |&gt; Sourceror . parse_string! ( ) ...&gt; |&gt; Sourceror . get_range ( include_comments : true ) %{ start : [ line : 1 , column : 1 ] , end : [ line : 2 , column : 11 ] }","ref":"Sourceror.html#get_range/2","title":"Sourceror.get_range/2","type":"function"},{"doc":"Returns the start position of a node. iex&gt; quoted = Sourceror . parse_string! ( &quot; :foo&quot; ) iex&gt; Sourceror . get_start_position ( quoted ) [ line : 1 , column : 2 ] iex&gt; quoted = Sourceror . parse_string! ( &quot; \\n \\n foo()&quot; ) iex&gt; Sourceror . get_start_position ( quoted ) [ line : 3 , column : 1 ] iex&gt; quoted = Sourceror . parse_string! ( &quot;Foo.{Bar}&quot; ) iex&gt; Sourceror . get_start_position ( quoted ) [ line : 1 , column : 1 ] iex&gt; quoted = Sourceror . parse_string! ( &quot;foo[:bar]&quot; ) iex&gt; Sourceror . get_start_position ( quoted ) [ line : 1 , column : 1 ] iex&gt; quoted = Sourceror . parse_string! ( &quot;foo(:bar)&quot; ) iex&gt; Sourceror . get_start_position ( quoted ) [ line : 1 , column : 1 ]","ref":"Sourceror.html#get_start_position/2","title":"Sourceror.get_start_position/2","type":"function"},{"doc":"Parses a single expression from the given string. It tries to parse on a per-line basis. Returns {:ok, quoted, rest} on success or {:error, source} on error. Examples iex&gt; ~S&quot;&quot;&quot; ...&gt; 42 ...&gt; ...&gt; :ok ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_expression ( ) { :ok , { :__block__ , [ trailing_comments : [ ] , leading_comments : [ ] , token : &quot;42&quot; , line : 2 , column : 1 ] , [ 42 ] } , &quot; \\n :ok&quot; } Options :from_line - The line at where the parsing should start. Defaults to 1 .","ref":"Sourceror.html#parse_expression/2","title":"Sourceror.parse_expression/2","type":"function"},{"doc":"Parses the source code into an extended AST suitable for source manipulation as described in Code.quoted_to_algebra/2 . Two additional fields are added to nodes metadata: :leading_comments - a list holding the comments found before the node. :trailing_comments - a list holding the comments found before the end of the node. For example, comments right before the end keyword. Comments are the same maps returned by Code.string_to_quoted_with_comments/2 .","ref":"Sourceror.html#parse_string/1","title":"Sourceror.parse_string/1","type":"function"},{"doc":"Same as parse_string/1 but raises on error.","ref":"Sourceror.html#parse_string!/1","title":"Sourceror.parse_string!/1","type":"function"},{"doc":"Applies one or more patches to the given string. A patch is a map containing the following values: :range - a map containing :start and :end keys, whose values are keyword lists containing the :line and :column representing the boundary of the patch. :change - the string being patched in or a function that takes the text of the patch range and returns the replacement. :preserve_indentation (default true ) - whether to automatically correct the indentation of the patch string to preserve the indentation level of the patch range (see examples below) Note that :line and :column start at 1 and represent a cursor positioned before the targeted position. For instance, here's how you would select the string &quot;ToBePatched&quot; in the following example: defmodule ToBePatched do # ^ ^ # col 11 22 %{ range : %{ start : [ line : 1 , column : 11 ] , end : [ line : 1 , column : 22 ] } } Ranges are usually derived from parsed AST nodes. See get_range/2 for more. Patches are applied bottom-up, such that patches to the beginning of the string do not interfere with the line/column of patches that come later. However, no checks are done for overlapping ranges, so take care to pass in non-overlapping patches. Examples iex&gt; original = ~S&quot;&quot;&quot; ...&gt; if not allowed? do ...&gt; raise &quot;Not allowed!&quot; ...&gt; end ...&gt; &quot;&quot;&quot; iex&gt; patch = %{ ...&gt; change : &quot;unless allowed? do \\n raise \\&quot; Not allowed! \\&quot; \\n end&quot; , ...&gt; range : %{ start : [ line : 1 , column : 1 ] , end : [ line : 3 , column : 4 ] } ...&gt; } iex&gt; Sourceror . patch_string ( original , [ patch ] ) ~S&quot;&quot;&quot; unless allowed? do raise &quot;Not allowed!&quot; end &quot;&quot;&quot; A range can also be a function, in which case the original text in the patch range will be given as an argument: iex&gt; original = ~S&quot;&quot;&quot; ...&gt; hello :world ...&gt; &quot;&quot;&quot; iex&gt; patch = %{ ...&gt; change : &amp; String . upcase / 1 , ...&gt; range : %{ start : [ line : 1 , column : 7 ] , end : [ line : 1 , column : 13 ] } ...&gt; } iex&gt; Sourceror . patch_string ( original , [ patch ] ) ~S&quot;&quot;&quot; hello :WORLD &quot;&quot;&quot; By default, lines after the first line of the patch will be indented relative to the indentation level at the start of the range: iex&gt; original = ~S&quot;&quot;&quot; ...&gt; outer do ...&gt; inner(foo do ...&gt; :original ...&gt; end) ...&gt; end ...&gt; &quot;&quot;&quot; iex&gt; patch = %{ ...&gt; change : &quot;bar do \\n :replacement \\n end&quot; , ...&gt; range : %{ start : [ line : 2 , column : 9 ] , end : [ line : 4 , column : 6 ] } ...&gt; } iex&gt; Sourceror . patch_string ( original , [ patch ] ) ~S&quot;&quot;&quot; outer do inner(bar do :replacement end) end &quot;&quot;&quot; If you don't want this behavior, you can add :preserve_indentation: false to your patch: iex&gt; original = ~S&quot;&quot;&quot; ...&gt; outer do ...&gt; inner(foo do ...&gt; :original ...&gt; end) ...&gt; end ...&gt; &quot;&quot;&quot; iex&gt; patch = %{ ...&gt; change : &quot;bar do \\n :replacement \\n end&quot; , ...&gt; range : %{ start : [ line : 2 , column : 9 ] , end : [ line : 4 , column : 6 ] } , ...&gt; preserve_indentation : false ...&gt; } iex&gt; Sourceror . patch_string ( original , [ patch ] ) ~S&quot;&quot;&quot; outer do inner(bar do :replacement end) end &quot;&quot;&quot;","ref":"Sourceror.html#patch_string/2","title":"Sourceror.patch_string/2","type":"function"},{"doc":"Performs a depth-first post-order traversal of a quoted expression. See postwalk/3 for more information.","ref":"Sourceror.html#postwalk/2","title":"Sourceror.postwalk/2","type":"function"},{"doc":"Performs a depth-first post-order traversal of a quoted expression with an accumulator. fun is a function that will receive the current node as a first argument and the traversal state as the second one. It must return a {quoted, state} , in the same way it would return {quoted, acc} when using Macro.postwalk/3 . The state is a map with the following keys: :acc - The accumulator. Defaults to nil if none is given.","ref":"Sourceror.html#postwalk/3","title":"Sourceror.postwalk/3","type":"function"},{"doc":"Prepends comments to the leading or trailing comments of a node.","ref":"Sourceror.html#prepend_comments/3","title":"Sourceror.prepend_comments/3","type":"function"},{"doc":"Performs a depth-first pre-order traversal of a quoted expression. See prewalk/3 for more information.","ref":"Sourceror.html#prewalk/2","title":"Sourceror.prewalk/2","type":"function"},{"doc":"Performs a depth-first pre-order traversal of a quoted expression with an accumulator. fun is a function that will receive the current node as a first argument and the traversal state as the second one. It must return a {quoted, state} , in the same way it would return {quoted, acc} when using Macro.prewalk/3 . The state is a map with the following keys: :acc - The accumulator. Defaults to nil if none is given.","ref":"Sourceror.html#prewalk/3","title":"Sourceror.prewalk/3","type":"function"},{"doc":"A wrapper around Code.quoted_to_algebra/2 for compatibility with pre 1.13 Elixir versions.","ref":"Sourceror.html#quoted_to_algebra/2","title":"Sourceror.quoted_to_algebra/2","type":"macro"},{"doc":"A wrapper around Code.string_to_quoted_with_comments/2 for compatibility with pre 1.13 Elixir versions.","ref":"Sourceror.html#string_to_quoted/2","title":"Sourceror.string_to_quoted/2","type":"macro"},{"doc":"A wrapper around Code.string_to_quoted_with_comments!/2 for compatibility with pre 1.13 Elixir versions.","ref":"Sourceror.html#string_to_quoted!/2","title":"Sourceror.string_to_quoted!/2","type":"macro"},{"doc":"Converts a quoted expression to a string. The comments line number will be ignored and the line number of the associated node will be used when formatting the code. If you're using Elixir &gt;=1.13, the locals_without_parens from your project formatter configuration will be used. You can pass a different set of options by using the locals_without_parens option. If you want to disable that option entirely, use locals_without_parens: [] . Options :indent - how many indentations to insert at the start of each line. Note that this only prepends the indents without checking the indentation of nested blocks. Defaults to 0 . :indent_type - the type of indentation to use. It can be one of :spaces , :single_space or :tabs . Defaults to :spaces . :format - if set to :splicing , if the quoted expression is a list, it will strip the square brackets. This is useful to print a single element of a keyword list. :quoted_to_algebra - expects a function of the type (Macro.t(), keyword -&gt; Inspect.Algebra.t()) to convert the given quoted expression to an algebra document. For more options see Code.format_string!/1 and Code.quoted_to_algebra/2 .","ref":"Sourceror.html#to_string/2","title":"Sourceror.to_string/2","type":"function"},{"doc":"Updates the arguments for the given node. iex&gt; node = { :foo , [ line : 1 ] , [ { :__block__ , [ line : 1 ] , [ 2 ] } ] } iex&gt; updater = fn args -&gt; Enum . map ( args , &amp; Sourceror . correct_lines ( &amp;1 , 2 ) ) end iex&gt; Sourceror . update_args ( node , updater ) { :foo , [ line : 1 ] , [ { :__block__ , [ line : 3 ] , [ 2 ] } ] }","ref":"Sourceror.html#update_args/2","title":"Sourceror.update_args/2","type":"function"},{"doc":"","ref":"Sourceror.html#t:comment/0","title":"Sourceror.comment/0","type":"type"},{"doc":"","ref":"Sourceror.html#t:patch/0","title":"Sourceror.patch/0","type":"type"},{"doc":"","ref":"Sourceror.html#t:position/0","title":"Sourceror.position/0","type":"type"},{"doc":"","ref":"Sourceror.html#t:range/0","title":"Sourceror.range/0","type":"type"},{"doc":"","ref":"Sourceror.html#t:traversal_function/0","title":"Sourceror.traversal_function/0","type":"type"},{"doc":"Utilities to merge an un-merge comments and quoted expressions.","ref":"Sourceror.Comments.html","title":"Sourceror.Comments","type":"module"},{"doc":"Does the opposite of merge_comments/2 , it extracts the comments from the quoted expression and returns both as a {quoted, comments} tuple.","ref":"Sourceror.Comments.html#extract_comments/2","title":"Sourceror.Comments.extract_comments/2","type":"function"},{"doc":"Merges the comments into the given quoted expression. The comments are inserted into the metadata of their closest node. Comments in the same line of before a node are inserted into the :leading_comments field while comments that are right before an end keyword are inserted into the :trailing_comments field.","ref":"Sourceror.Comments.html#merge_comments/2","title":"Sourceror.Comments.merge_comments/2","type":"function"},{"doc":"Functions to identify an classify forms and quoted expressions.","ref":"Sourceror.Identifier.html","title":"Sourceror.Identifier","type":"module"},{"doc":"Checks if the given quoted form is an atomic literal in the AST. This set includes numbers, atoms, and strings, but not collections like tuples, lists, or maps. This guard returns true for literals that are the only elements inside of a :__block__ , such as {:__block__, [], [:literal]} . Examples iex&gt; is_atomic_literal ( 1 ) true iex&gt; is_atomic_literal ( 1.0 ) true iex&gt; is_atomic_literal ( :foo ) true iex&gt; is_atomic_literal ( &quot;foo&quot; ) true iex&gt; is_atomic_literal ( { :__block__ , [ ] , [ 1 ] } ) true iex&gt; is_atomic_literal ( { :__block__ , [ ] , [ 1 , 2 ] } ) false iex&gt; is_atomic_literal ( { :__block__ , [ ] , [ { :node , [ ] , nil } ] } ) false iex&gt; is_atomic_literal ( &#39;foo&#39; ) false","ref":"Sourceror.Identifier.html#is_atomic_literal/1","title":"Sourceror.Identifier.is_atomic_literal/1","type":"macro"},{"doc":"Checks if the given identifier is a binary op. Examples iex&gt; is_binary_op ( :+ ) true","ref":"Sourceror.Identifier.html#is_binary_op/1","title":"Sourceror.Identifier.is_binary_op/1","type":"macro"},{"doc":"Checks if the given quoted form is a call. Calls are any form of the shape {form, metadata, args} where args is a list, with the exception of blocks and aliases, which are identified by the forms :__block__ and :__aliases__ . Examples iex&gt; &quot;node()&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_call ( ) true iex&gt; &quot;Kernel.node()&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_call ( ) true iex&gt; &quot;%{}&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_call ( ) true iex&gt; &quot;@attr&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_call ( ) true iex&gt; &quot;node&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_call ( ) false iex&gt; &quot;1&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_call ( ) false iex&gt; &quot;(1; 2)&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_call ( ) false iex&gt; &quot;Macro.Env&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_call ( ) false","ref":"Sourceror.Identifier.html#is_call/1","title":"Sourceror.Identifier.is_call/1","type":"macro"},{"doc":"Checks if the given quoted form is an identifier, such as a variable. Examples iex&gt; &quot;node&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_identifier ( ) true iex&gt; &quot;node()&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_identifier ( ) false iex&gt; &quot;1&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_identifier ( ) false","ref":"Sourceror.Identifier.html#is_identifier/1","title":"Sourceror.Identifier.is_identifier/1","type":"macro"},{"doc":"Checks if the given identifier is a pipeline operator. Examples iex&gt; is_pipeline_op ( :|&gt; ) true","ref":"Sourceror.Identifier.html#is_pipeline_op/1","title":"Sourceror.Identifier.is_pipeline_op/1","type":"macro"},{"doc":"Checks if the given quoted form is a qualified call. All unqualified calls would also return true if passed to is_call/1 , but they have the shape {{:., dot_metadata, dot_args}, metadata, args} . Examples iex&gt; &quot;Kernel.node()&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_qualified_call ( ) true iex&gt; &quot;__MODULE__.node()&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_qualified_call ( ) true iex&gt; &quot;foo.()&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_qualified_call ( ) true iex&gt; &quot;foo.bar()&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_qualified_call ( ) true iex&gt; &quot;node()&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_qualified_call ( ) false iex&gt; &quot;%{}&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_qualified_call ( ) false iex&gt; &quot;@attr&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_qualified_call ( ) false iex&gt; &quot;1&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_qualified_call ( ) false iex&gt; &quot;(1; 2)&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_qualified_call ( ) false iex&gt; &quot;Macro.Env&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_qualified_call ( ) false","ref":"Sourceror.Identifier.html#is_qualified_call/1","title":"Sourceror.Identifier.is_qualified_call/1","type":"macro"},{"doc":"Checks if the given identifier is an unary op. Examples iex&gt; is_unary_op ( :+ ) true","ref":"Sourceror.Identifier.html#is_unary_op/1","title":"Sourceror.Identifier.is_unary_op/1","type":"macro"},{"doc":"Checks if the given quoted form is an unqualified call. All unqualified calls would also return true if passed to is_call/1 , but they have the shape {atom, metadata, args} . Examples iex&gt; &quot;node()&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_unqualified_call ( ) true iex&gt; &quot;%{}&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_unqualified_call ( ) true iex&gt; &quot;@attr&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_unqualified_call ( ) true iex&gt; &quot;node&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_unqualified_call ( ) false iex&gt; &quot;1&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_unqualified_call ( ) false iex&gt; &quot;(1; 2)&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_unqualified_call ( ) false iex&gt; &quot;Macro.Env&quot; |&gt; Sourceror . parse_string! ( ) |&gt; is_unqualified_call ( ) false","ref":"Sourceror.Identifier.html#is_unqualified_call/1","title":"Sourceror.Identifier.is_unqualified_call/1","type":"macro"},{"doc":"Checks if the given atom is a valid module alias. Examples iex&gt; valid_alias? ( Foo ) true iex&gt; valid_alias? ( :foo ) false","ref":"Sourceror.Identifier.html#valid_alias?/1","title":"Sourceror.Identifier.valid_alias?/1","type":"function"},{"doc":"Functions that generate patches for common operations. Functions in this module assume that the AST was parsed using Sourceror functions and that it wasn't modified. If you changed the tree before calling Sourceror.Patch functions, then the patch ranges are not guaranteed to match 1:1 with the original source code.","ref":"Sourceror.Patch.html","title":"Sourceror.Patch","type":"module"},{"doc":"Renames a qualified or unqualified function call. iex&gt; original = &quot;String.to_atom(foo)&quot; iex&gt; ast = Sourceror . parse_string! ( original ) iex&gt; patches = Sourceror.Patch . rename_call ( ast , :to_existing_atom ) iex&gt; Sourceror . patch_string ( original , patches ) &quot;String.to_existing_atom(foo)&quot; If the call is a sigil, you only need to provide the replacement letter: iex&gt; original = &quot;~H(foo)&quot; iex&gt; ast = Sourceror . parse_string! ( original ) iex&gt; patches = Sourceror.Patch . rename_call ( ast , :F ) iex&gt; Sourceror . patch_string ( original , patches ) &quot;~F(foo)&quot;","ref":"Sourceror.Patch.html#rename_call/2","title":"Sourceror.Patch.rename_call/2","type":"function"},{"doc":"Renames an identifier(ie a variable name). Examples iex&gt; original = &quot;foo&quot; iex&gt; ast = Sourceror . parse_string! ( original ) iex&gt; patches = Sourceror.Patch . rename_identifier ( ast , :bar ) iex&gt; Sourceror . patch_string ( original , patches ) &quot;bar&quot;","ref":"Sourceror.Patch.html#rename_identifier/2","title":"Sourceror.Patch.rename_identifier/2","type":"function"},{"doc":"Generates patches that rename the keys of a keyword list. The replacements is a keyword list, with the keys to replace as keys, and the replacement as the value. Examples iex&gt; original = &quot;[a: b, c: d, e: f]&quot; iex&gt; ast = Sourceror . parse_string! ( original ) iex&gt; patches = Sourceror.Patch . rename_kw_keys ( ast , a : :foo , e : :bar ) iex&gt; Sourceror . patch_string ( original , patches ) &quot;[foo: b, c: d, bar: f]&quot;","ref":"Sourceror.Patch.html#rename_kw_keys/2","title":"Sourceror.Patch.rename_kw_keys/2","type":"function"},{"doc":"The state struct for Sourceror traversal functions.","ref":"Sourceror.TraversalState.html","title":"Sourceror.TraversalState","type":"module"},{"doc":"","ref":"Sourceror.TraversalState.html#t:t/0","title":"Sourceror.TraversalState.t/0","type":"type"},{"doc":"Tree-like data structure that provides enhanced navigation and modification of an Elixir AST. This implementation is based on Gérard Huet Functional pearl: the zipper and Clojure's clojure.zip API. A zipper is a data structure that represents a location in a tree from the perspective of the current node, also called the focus . It is represented by a struct containing a :node and :path , which is a private field used to track the position of the :node with regards to the entire tree. The :path is an implementation detail that should be considered private. For more information and examples, see the following guides: Zippers (an introduction) Expand multi-alias syntax (an example)","ref":"Sourceror.Zipper.html","title":"Sourceror.Zipper","type":"module"},{"doc":"Inserts the item as the rightmost child of the node at this zipper, without moving.","ref":"Sourceror.Zipper.html#append_child/2","title":"Sourceror.Zipper.append_child/2","type":"function"},{"doc":"","ref":"Sourceror.Zipper.html#branch?/1","title":"Sourceror.Zipper.branch?/1","type":"function"},{"doc":"Returns a list of children of the node.","ref":"Sourceror.Zipper.html#children/1","title":"Sourceror.Zipper.children/1","type":"function"},{"doc":"Returns the zipper of the leftmost child of the node at this zipper, or nil if no there's no children.","ref":"Sourceror.Zipper.html#down/1","title":"Sourceror.Zipper.down/1","type":"function"},{"doc":"Returns a zipper to the node that satisfies the predicate function, or nil if none is found. The optional second parameters specifies the direction , defaults to :next .","ref":"Sourceror.Zipper.html#find/3","title":"Sourceror.Zipper.find/3","type":"function"},{"doc":"Inserts the item as the leftmost child of the node at this zipper, without moving.","ref":"Sourceror.Zipper.html#insert_child/2","title":"Sourceror.Zipper.insert_child/2","type":"function"},{"doc":"Inserts the item as the left sibling of the node at this zipper, without moving. Raises an ArgumentError when attempting to insert a sibling at the top level.","ref":"Sourceror.Zipper.html#insert_left/2","title":"Sourceror.Zipper.insert_left/2","type":"function"},{"doc":"Inserts the item as the right sibling of the node at this zipper, without moving. Raises an ArgumentError when attempting to insert a sibling at the top level.","ref":"Sourceror.Zipper.html#insert_right/2","title":"Sourceror.Zipper.insert_right/2","type":"function"},{"doc":"Returns the zipper of the left sibling of the node at this zipper, or nil.","ref":"Sourceror.Zipper.html#left/1","title":"Sourceror.Zipper.left/1","type":"function"},{"doc":"Returns the leftmost sibling of the node at this zipper, or itself.","ref":"Sourceror.Zipper.html#leftmost/1","title":"Sourceror.Zipper.leftmost/1","type":"function"},{"doc":"Returns a new branch node, given an existing node and new children.","ref":"Sourceror.Zipper.html#make_node/2","title":"Sourceror.Zipper.make_node/2","type":"function"},{"doc":"Returns the following zipper in depth-first pre-order.","ref":"Sourceror.Zipper.html#next/1","title":"Sourceror.Zipper.next/1","type":"function"},{"doc":"Returns the node at the zipper.","ref":"Sourceror.Zipper.html#node/1","title":"Sourceror.Zipper.node/1","type":"function"},{"doc":"Returns the previous zipper in depth-first pre-order. If it's already at the end, it returns nil.","ref":"Sourceror.Zipper.html#prev/1","title":"Sourceror.Zipper.prev/1","type":"function"},{"doc":"Removes the node at the zipper, returning the zipper that would have preceded it in a depth-first walk.","ref":"Sourceror.Zipper.html#remove/1","title":"Sourceror.Zipper.remove/1","type":"function"},{"doc":"Replaces the current node in the zipper with a new node.","ref":"Sourceror.Zipper.html#replace/2","title":"Sourceror.Zipper.replace/2","type":"function"},{"doc":"Returns the zipper of the right sibling of the node at this zipper, or nil.","ref":"Sourceror.Zipper.html#right/1","title":"Sourceror.Zipper.right/1","type":"function"},{"doc":"Returns the rightmost sibling of the node at this zipper, or itself.","ref":"Sourceror.Zipper.html#rightmost/1","title":"Sourceror.Zipper.rightmost/1","type":"function"},{"doc":"Walks the zipper all the way up and returns the root node.","ref":"Sourceror.Zipper.html#root/1","title":"Sourceror.Zipper.root/1","type":"function"},{"doc":"Returns the zipper of the right sibling of the node at this zipper, or the next zipper when no right sibling is available. This allows to skip subtrees while traversing the siblings of a node. The optional second parameters specifies the direction , defaults to :next . If no right/left sibling is available, this function returns the same value as next/1 / prev/1 . The function skip/1 behaves like the :skip in traverse_while/2 and traverse_while/3 .","ref":"Sourceror.Zipper.html#skip/2","title":"Sourceror.Zipper.skip/2","type":"function"},{"doc":"Walks the zipper all the way up and returns the top zipper.","ref":"Sourceror.Zipper.html#top/1","title":"Sourceror.Zipper.top/1","type":"function"},{"doc":"Traverses the tree in depth-first pre-order calling the given function for each node. When the traversal is finished, the zipper will be back where it began. If the zipper is not at the top, just the subtree will be traversed. The function must return a zipper.","ref":"Sourceror.Zipper.html#traverse/2","title":"Sourceror.Zipper.traverse/2","type":"function"},{"doc":"Traverses the tree in depth-first pre-order calling the given function for each node with an accumulator. When the traversal is finished, the zipper will be back where it began. If the zipper is not at the top, just the subtree will be traversed.","ref":"Sourceror.Zipper.html#traverse/3","title":"Sourceror.Zipper.traverse/3","type":"function"},{"doc":"Traverses the tree in depth-first pre-order calling the given function for each node. The traversing will continue if the function returns {:cont, zipper} , skipped for {:skip, zipper} and halted for {:halt, zipper} . When the traversal is finished, the zipper will be back where it began. If the zipper is not at the top, just the subtree will be traversed. The function must return a zipper.","ref":"Sourceror.Zipper.html#traverse_while/2","title":"Sourceror.Zipper.traverse_while/2","type":"function"},{"doc":"Traverses the tree in depth-first pre-order calling the given function for each node with an accumulator. When the traversal is finished, the zipper will be back where it began. The traversing will continue if the function returns {:cont, zipper, acc} , skipped for {:skip, zipper, acc} and halted for {:halt, zipper, acc} If the zipper is not at the top, just the subtree will be traversed.","ref":"Sourceror.Zipper.html#traverse_while/3","title":"Sourceror.Zipper.traverse_while/3","type":"function"},{"doc":"Returns the zipper of the parent of the node at this zipper, or nil if at the top.","ref":"Sourceror.Zipper.html#up/1","title":"Sourceror.Zipper.up/1","type":"function"},{"doc":"Replaces the current node in the zipper with the result of applying fun to the node.","ref":"Sourceror.Zipper.html#update/2","title":"Sourceror.Zipper.update/2","type":"function"},{"doc":"Creates a zipper from a tree node.","ref":"Sourceror.Zipper.html#zip/1","title":"Sourceror.Zipper.zip/1","type":"function"},{"doc":"","ref":"Sourceror.Zipper.html#t:path/0","title":"Sourceror.Zipper.path/0","type":"opaque"},{"doc":"","ref":"Sourceror.Zipper.html#t:t/0","title":"Sourceror.Zipper.t/0","type":"type"},{"doc":"","ref":"Sourceror.Zipper.html#t:tree/0","title":"Sourceror.Zipper.tree/0","type":"type"},{"doc":"Provides Sourceror.Zipper 's implementation for the Inspect protocol. When inspecting a zipper, the default representation shows the current node and provides indicators of the surrounding context without displaying the full path, which can be very verbose for large ASTs. iex&gt; alias Sourceror.Zipper , as : Z Sourceror.Zipper iex&gt; code = &quot;&quot;&quot; ...&gt; def my_function do ...&gt; :ok ...&gt; end \\ ...&gt; &quot;&quot;&quot; iex&gt; zipper = code |&gt; Code . string_to_quoted! ( ) |&gt; Z . zip ( ) # Sourceror.Zipper &lt; #root { :def , [ line : 1 ] , [ { :my_function , [ line : 1 ] , nil } , [ do : :ok ] ] } &gt; iex&gt; zipper |&gt; Z . next ( ) # Sourceror.Zipper &lt; { :my_function , [ line : 1 ] , nil } #... &gt; This representation can be changed using default_inspect_as/1 to set a global default or by using :custom_options with the :zippers key when inspecting. (See Inspect.Opts for more about inspect options.) Zippers can be inspected in these formats: :as_ast (default, seen above) - display the current node as an AST :as_code - display the current node formatted as code :raw - display the raw %Sourceror.Zipper{} struct including the :path . Using the zipper defined above as an example: iex&gt; zipper |&gt; inspect ( custom_options : [ zipper : :as_code ] ) |&gt; IO . puts ( ) # Sourceror.Zipper &lt; #root def my_function do :ok end &gt; iex&gt; zipper |&gt; Z . next ( ) |&gt; inspect ( custom_options : [ zipper : :as_code ] ) |&gt; IO . puts ( ) # Sourceror.Zipper &lt; my_function #... &gt; iex&gt; zipper |&gt; Z . next ( ) |&gt; inspect ( custom_options : [ zipper : :raw ] , pretty : true ) |&gt; IO . puts ( ) % Sourceror.Zipper { node : { :my_function , [ line : 1 ] , nil } , path : %{ parent : % Sourceror.Zipper { node : { :def , [ line : 1 ] , [ { :my_function , [ line : 1 ] , nil } , [ do : :ok ] ] } , path : nil } , left : nil , right : [ [ do : :ok ] ] } }","ref":"Sourceror.Zipper.Inspect.html","title":"Sourceror.Zipper.Inspect","type":"module"},{"doc":"Sets the default inspection format for zippers. Examples Consider the following zipper: iex&gt; zipper # Sourceror.Zipper &lt; #root { :def , [ line : 1 ] , [ { :my_function , [ line : 1 ] , nil } , [ do : :ok ] ] } &gt; iex&gt; Sourceror.Zipper.Inspect . default_inspect_as ( :as_code ) :ok iex&gt; zipper # Sourceror.Zipper &lt; #root def my_function do :ok end &gt; iex&gt; Sourceror.Zipper.Inspect . default_inspect_as ( :raw ) :ok iex&gt; zipper % Sourceror.Zipper { node : { :def , [ line : 1 ] , [ { :my_function , [ line : 1 ] , nil } , [ do : :ok ] ] } , path : nil }","ref":"Sourceror.Zipper.Inspect.html#default_inspect_as/1","title":"Sourceror.Zipper.Inspect.default_inspect_as/1","type":"function"},{"doc":"Inspection formats for zippers.","ref":"Sourceror.Zipper.Inspect.html#t:inspect_as/0","title":"Sourceror.Zipper.Inspect.inspect_as/0","type":"type"},{"doc":"All notable changes to this project will be documented in this file. The format is based on Keep a Changelog , and this project adheres to Semantic Versioning .","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"1. Bug fixes [Sourceror] Handle unqualified double-parenthesis calls in get_range by @zachallaun in https://github.com/doorgan/sourceror/pull/111","ref":"changelog.html#v0-1410-2023-11-06","title":"Changelog - v0.1410 (2023-11-06)","type":"extras"},{"doc":"1. Enhancements [Sourceror] Change get_range/1 to allow returning nil by @zachallaun in https://github.com/doorgan/sourceror/pull/107parse_string` by @zachallaun in https://github.com/doorgan/sourceror/pull/104 - [Sourceror.Identifier] Add additional identification guards by @zachallaun in https://github.com/doorgan/sourceror/pull/108 ### 2. Bug fixes - [Sourceror] Ensure get_range works for all syntax nodes parsed using `Sourceror.","ref":"changelog.html#v0-14-0-2023-09-19","title":"Changelog - v0.14.0 (2023-09-19)","type":"extras"},{"doc":"The changelog for v0.13 releases can be found in the 0.13 branch .","ref":"changelog.html#v0-13-0-2023-08-23","title":"Changelog - v0.13.0 (2023-08-23)","type":"extras"},{"doc":"","ref":"contributing.html","title":"Contributing","type":"extras"},{"doc":"For issues like bug reports please file an issue in the Issue tracker .","ref":"contributing.html#issues","title":"Contributing - Issues","type":"extras"},{"doc":"To propose new features to the library, please open a thread in the Discussions section under the Proposals category. Development Contributions Take a look at the open issues here and post that you plan on working on an issue, so we can discuss approaches and make sure there isn't duplicated work happening. Before submitting a pull request, make sure all of the following pass: mix deps.get mix check","ref":"contributing.html#proposals","title":"Contributing - Proposals","type":"extras"},{"doc":"Apache License Version 2.0, January 2004 http://www.apache.org/licenses/ TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION 1. Definitions. &quot;License&quot; shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document. &quot;Licensor&quot; shall mean the copyright owner or entity authorized by the copyright owner that is granting the License. &quot;Legal Entity&quot; shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, &quot;control&quot; means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity. &quot;You&quot; (or &quot;Your&quot;) shall mean an individual or Legal Entity exercising permissions granted by this License. &quot;Source&quot; form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files. &quot;Object&quot; form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types. &quot;Work&quot; shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below). &quot;Derivative Works&quot; shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof. &quot;Contribution&quot; shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, &quot;submitted&quot; means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as &quot;Not a Contribution.&quot; &quot;Contributor&quot; shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work. 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form. 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed. 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions: (a) You must give any other recipients of the Work or Derivative Works a copy of this License; and (b) You must cause any modified files to carry prominent notices stating that You changed the files; and (c) You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and (d) If the Work includes a &quot;NOTICE&quot; text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License. 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions. 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file. 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License. 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages. 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. END OF TERMS AND CONDITIONS","ref":"license.html","title":"License","type":"extras"},{"doc":"Utilities to work with Elixir source code. NOTICE: This library is under heavy development. Expect frequent breaking changes until the first stable v1.0 release is out.","ref":"readme.html","title":"Overview","type":"extras"},{"doc":"Add :sourceror as a dependency to your project's mix.exs : defp deps do [ { :sourceror , &quot;~&gt; 0.14&quot; } ] end A note on compatibility Sourceror is compatible with Elixir versions down to 1.10 and OTP 21. For Elixir versions prior to 1.13 it uses a vendored version of the Elixir parser and formatter modules. This means that for Elixir versions prior to 1.12 it will successfully parse the new syntax for stepped ranges instead of raising a SyntaxError , but everything else should work as expected.","ref":"readme.html#installation","title":"Overview - Installation","type":"extras"},{"doc":"Be as close as possible to the standard Elixir AST. Make working with comments as simple as possible. No dev/prod dependencies, to simplify integration with other tools.","ref":"readme.html#goals-of-the-library","title":"Overview - Goals of the library","type":"extras"},{"doc":"Having the AST and comments as separate entities allows Elixir to expose the code formatting utilities without making any changes to it's AST, but also delegates the task of figuring out what's the most appropriate way to work with them to us. Sourceror's take is to use the node metadata to store the comments. This allows us to work with an AST that is as close to regular elixir AST as possible. It also allows you to move nodes around without worrying about leaving a comment behind and ending up with misplaced comments. Two metadata fields are added to the regular Elixir AST: :leading_comments - holds the comments directly above the node or are in the same line as it. For example: test &quot;parses leading comments&quot; do quoted = &quot;&quot;&quot; # Comment for :a :a # Also a comment for :a &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) assert { :__block__ , meta , [ :a ] } = quoted assert meta [ :leading_comments ] == [ %{ line : 1 , column : 1 , previous_eol_count : 1 , next_eol_count : 1 , text : &quot;# Comment for :a&quot; } , %{ line : 2 , column : 4 , previous_eol_count : 0 , next_eol_count : 1 , text : &quot;# Also a comment for :a&quot; } , ] end :trailing_comments - holds the comments that are inside of the node, but aren't leading any children, for example: test &quot;parses trailing comments&quot; do quoted = &quot;&quot;&quot; def foo() do :ok # A trailing comment end # Not a trailing comment for :foo &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) assert { :__block__ , block_meta , [ { :def , meta , _ } ] } = quoted assert [ %{ line : 3 , text : &quot;# A trailing comment&quot; } ] = meta [ :trailing_comments ] assert [ %{ line : 4 , text : &quot;# Not a trailing comment for :foo&quot; } ] = block_meta [ :trailing_comments ] end Note that Sourceror considers leading comments to the ones that are found in the same line as a node, and trailing comments to the ones that are found before the ending line of a node, based on the end , closing or end_of_expression line. This also makes the Sourceror AST consistent with the way the Elixir formatter works, making it easier to reason about how a given AST would be formatted.","ref":"readme.html#sourceror-s-ast","title":"Overview - Sourceror's AST","type":"extras"},{"doc":"Elixir provides the Macro.prewalk , Macro.postwalk and Macro.traverse functions to traverse the AST. You can use the same functions to traverse the Sourceror AST as well, since it has the same shape as the standard Elixir AST. Sourceror also provides the Sourceror.prewalk , Sourceror.postwalk and Sourceror.traverse variants. At the time of writing they are mostly wrappers around the standard Elixir functions for AST traversal, but they may be enhanced in the future if more AST formats are introduced. In addition to these, Sourceror also provides a Zipper implementation for the Elixir AST. You can learn more about it in the Zippers notebook .","ref":"readme.html#traversing-the-ast","title":"Overview - Traversing the AST","type":"extras"},{"doc":"You can use Sourceror to manipulate the AST and turn it back into human readable Elixir code, this is commonly known as writing a &quot;codemod&quot;. For example, you can write a codemod to replace calls to String.to_atom to String.to_existing_atom : test &quot;updates the source code&quot; do source = &quot;&quot;&quot; String.to_atom(foo) \\ &quot;&quot;&quot; new_source = source |&gt; Sourceror . parse_string! ( ) |&gt; Macro . postwalk ( fn { { :. , dot_meta , [ { :__aliases__ , alias_meta , [ :String ] } , :to_atom ] } , call_meta , args } -&gt; { { :. , dot_meta , [ { :__aliases__ , alias_meta , [ :String ] } , :to_existing_atom ] } , call_meta , args } quoted -&gt; quoted end ) |&gt; Sourceror . to_string ( ) assert new_source == &quot;&quot;&quot; String.to_existing_atom(foo) \\ &quot;&quot;&quot; end However, this will affect the whole source code, as we are working on the full source AST. Sourceror relies on the Elixir formatter to produce human readable code, so the original code formatting will be lost by using Sourceror.to_string . If your code is already using the Elixir formatter then this won't be an issue, but it will be an undesirable effect if you're not using it. An alternative to this is to use Patches instead. A patch is a data structure that specifies the text range that should be replaced, and either a replacement string or a function that takes the text in that range and produces a string replacement. Using patches, we could do the same as above, but produce a patch instead of modifying the AST. As a result, only the parts that need to be changed will be affected, and the rest of the code keeps the original formatting: test &quot;patches the source code&quot; do source = &quot;&quot;&quot; case foo do nil -&gt; :bar _ -&gt; String.to_atom(foo) end \\ &quot;&quot;&quot; { _quoted , patches } = source |&gt; Sourceror . parse_string! ( ) |&gt; Macro . postwalk ( [ ] , fn { { :. , dot_meta , [ { :__aliases__ , alias_meta , [ :String ] } , :to_atom ] } , call_meta , args } = quoted , patches -&gt; range = Sourceror . get_range ( quoted ) replacement = { { :. , dot_meta , [ { :__aliases__ , alias_meta , [ :String ] } , :to_existing_atom ] } , call_meta , args } |&gt; Sourceror . to_string ( ) patch = %{ range : range , change : replacement } { quoted , [ patch | patches ] } quoted , patches -&gt; { quoted , patches } end ) assert Sourceror . patch_string ( source , patches ) == &quot;&quot;&quot; case foo do nil -&gt; :bar _ -&gt; String.to_existing_atom(foo) end \\ &quot;&quot;&quot; end You have to keep in mind that: If you patch a node that has inner code, like replacing a full case , then the contents of the node will be reformatted as well. At the moment, Sourceror won't check for conflicts in the patches ranges, so care needs to be taken to not produce conflicting patches. You may need to do a number of parse -&gt; patch -&gt; reparse if you find yourself generating conflicting patches. Some of the most common patching operations are available in the Sourceror.Patch module","ref":"readme.html#patching-the-source-code","title":"Overview - Patching the source code","type":"extras"},{"doc":"There have been several attempts at source code manipulation in the Elixir community. Thanks to its metaprogramming features, Elixir provides builtin tools that let us get the AST of any Elixir code, but when it comes to turning the AST back to code as text, we had limited options. Macro.to_string/2 is a thing, but the produced code is generally ugly, mostly because of the extra parenthesis or because it turns string interpolations into calls to erlang modules, to name some examples. This meant that, even if we could use Macro.to_string/2 to get a string and then give that to the Elixir formatter Code.format_string!/2 , the output would still be suboptimal, as the formatter is not designed to change the semantics of the code, only to pretty print it. For example, call to erlang modules would be kept as is instead of being turned back to interpolations. We also had the additional problem of comments being discarded by the tokenizer, and literals not having information like line numbers or delimiter characters. This makes the regular AST too lossy to be useful if what we want is to manipulate the source code, because we need as much information as possible to be able to stay as close to the source as possible. There have been several proposal in the past to bring all this information to the Elixir AST, but they all meant a change that would either break macros due to the addition of new types of AST nodes, or making a compromise in core Elixir itself by storing comments in the nods metadata. This discussion in the Elixir mailing list highlights the various issues faced when deciding if and how the comments would be preserved. Arjan Scherpenisse also did a talk where he discusses about the problems of using the standard Elixir AST to build refactoring tools. Despite of all these issues, the Elixir formatter is still capable of manipulating the source code to pretty print it. Under the hood it does some neat tricks to have all this information available: on one hand, it tells the tokenizer to extract the comments from the source code and keep it at hand(not in the AST itself, but as a separate data structure), and on the other hand it tells the parser to wrap literals in block nodes so metadata can be preserved. Once it has all it needs, it can start converting the AST and comments into an algebra document, and ultimately convert that to a string. This functionality was private, and if we wanted to do it ourselves we would have to replicate or vendor the Elixir formatter with all its more than 2000 lines of code. This approach was explored by Wojtek Mach in wojtekmach/fix , but it involved vendoring the elixir Formatter code, was tightly coupled to the formatting process, and any change in Elixir would break the code. Since Elixir 1.13 this functionality from the formatter was finally exposed via the Code.string_to_quoted_with_comments/2 and Code.quoted_to_algebra/2 functions. The former gives us access to the list of comments in a shape the Elixir formatter is able to use, and the latter lets us turn any arbitrary Elixir AST into an algebra document. If we also give it the list of comments, it will merge them together, allowing us to format AST and preserve the comments. Now all we need to care about is of manipulating the AST, and let the formatter do the rest.","ref":"readme.html#background","title":"Overview - Background","type":"extras"},{"doc":"You can find Sourceror documentation on Hex Docs .","ref":"readme.html#documentation","title":"Overview - Documentation","type":"extras"},{"doc":"You can find usage examples in the examples folder. You can run them by cloning the repo and running elixir examples/&lt;example_file&gt;.exs . You can also find documented examples you can run with Livebook in the notebooks folder.","ref":"readme.html#examples","title":"Overview - Examples","type":"extras"},{"doc":"If you want to contribute to Sourceror, please check our Contributing section for pointers.","ref":"readme.html#contributing","title":"Overview - Contributing","type":"extras"},{"doc":"If you have any questions about Sourceror or need assistance, please open a thread in the Discussions section.","ref":"readme.html#getting-assistance","title":"Overview - Getting assistance","type":"extras"},{"doc":"Copyright (c) 2021 dorgandash@gmail.com Licensed under the Apache License, Version 2.0 (the &quot;License&quot;); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.","ref":"readme.html#copyright-and-license","title":"Overview - Copyright and License","type":"extras"},{"doc":"","ref":"zippers.html","title":"Zippers","type":"extras"},{"doc":"Elixir provides the functions Macro.traverse/4 , Macro.prewalk/3 and Macro.postwalk/3 to traverse the AST. They are very useful if we want to collect the nodes of some type or manipulate the current node we are visiting, but it becomes increasingly hard to use when we need to remove a node, insert siblings, or get information about other nodes. To mitigate these issues, we can reach for a zipper . A zipper is a data structure that represents a location in some data, composed of a focus (the current element we're visiting) and some context that holds the rest of the data from the perspective of that focus. To visualize this, we'll use a list of integers from 1 to 10 and say that our current focus is the number 5 : [ 1 , 2 , 3 , 4 , 5 , 6 , 7 , 8 , 9 , 10 ] ↑ focus From the perspective of the 5 , the surrounding elements are the numbers from 4 to 1 at it's left, and the numbers from 6 to 10 at it's right. We can encode that information in a tuple: focus ↓ { [ 4 , 3 , 2 , 1 ] , 5 , [ 6 , 7 , 8 , 9 , 10 ] } ↑ ↑ left siblings right siblings That data structure, composed of the focus and the rest of the data as viewed from the focus, is called a zipper . In a traditional list, if we wanted to go to the previous element, we would have to start from the beginning of the list and traverse it all the way down until the current element minus one position. With a list zipper, however, all we need to do is to pop the first element of the left siblings to get the new focus, and put the old focus in the list of right siblings. After that we get a new zipper focused on 4 : { [ 3 , 2 , 1 ] , 4 , [ 5 , 6 , 7 , 8 , 9 , 10 ] } If we wanted to move to the right, we would perform the same operation, but popping from the right and prepending to the left. Such movements are performed in constant time, making zippers an attractive option when a lot of movements over a list are required, but things start to get more interesting when we use zippers with trees. When moving across a list, we are constrained to horizontal movements(left and right), but in trees we have an extra dimension to move around(up and down). So we need to encode at least four things: the focus, the left and right siblings, and the parent nodes in the tree. Sourceror provides a zipper implementation for the Elixir AST, let's see what it looks like. First, we need to install Sourceror: Mix . install ( [ :sourceror ] ) The module that implements zippers for the Elixir AST is Sourceror.Zipper . We'll use the alias Z for convenience. alias Sourceror.Zipper , as : Z Sourceror.Zipper To create a zipper for a tree, we use Z.zip/1 . Let's first start with a tree of lists first, since they are easier to visualize than a full blown AST: zipper = Z . zip ( [ 1 , [ 2 , 3 ] , 4 , [ 5 , [ 6 , 7 ] ] , 8 ] ) # Sourceror.Zipper &lt; #root [ 1 , [ 2 , 3 ] , 4 , [ 5 , [ 6 , 7 ] ] , 8 ] &gt; Sourceror represents zippers using the %Sourceror.Zipper{} struct. When this zipper is inspected, the #root tag means we're at the topmost node in the tree. Let's see what happens when we go down a level: zipper |&gt; Z . down ( ) # Sourceror.Zipper &lt; 1 #... &gt; Now it's getting more interesting! The current node of the zipper is 1 , meaning we went down a level and are now focusing on the first element of the list. The #... tag after the 1 means that the current node has right siblings. If we move one position to the right, we get a different zipper: zipper |&gt; Z . down ( ) |&gt; Z . right ( ) # Sourceror.Zipper &lt; #... [ 2 , 3 ] #... &gt; Now we're focusing on [2, 3] , which has siblings both to its left (the 1 we saw previously) and to its right. Sourceror defines a less-verbose implementation of the Inspect protocol for zippers, but the default format hides some of the implementation details. To better understand the zipper's internals, we'll set a more verbose default format to use going forward. (You can check out the Sourceror.Zipper.Inspect docs for more info.) Sourceror.Zipper.Inspect . default_inspect_as ( :raw ) zipper |&gt; Z . down ( ) |&gt; Z . right ( ) % Sourceror.Zipper { node : [ 2 , 3 ] , path : %{ parent : % Sourceror.Zipper { node : [ 1 , [ 2 , 3 ] , 4 , [ 5 , [ 6 , 7 ] ] , 8 ] , path : nil } , left : [ 1 ] , right : [ 4 , [ 5 , [ 6 , 7 ] ] , 8 ] } } Now we can see that the zipper struct is composed of two fields, the current :node and a :path that contains siblings to the left, siblings to the right, and the parent. Besides moving around, we can also change the tree as a localized operation. For example, let's add an element to the right of our position: zipper |&gt; Z . down ( ) |&gt; Z . right ( ) |&gt; Z . insert_right ( :to_the_right! ) % Sourceror.Zipper { node : [ 2 , 3 ] , path : %{ parent : % Sourceror.Zipper { node : [ 1 , [ 2 , 3 ] , 4 , [ 5 , [ 6 , 7 ] ] , 8 ] , path : nil } , left : [ 1 ] , right : [ :to_the_right! , 4 , [ 5 , [ 6 , 7 ] ] , 8 ] } } We can see that the atom :to_the_right! was added to our right siblings. Let's now remove it: zipper |&gt; Z . down ( ) |&gt; Z . right ( ) |&gt; Z . insert_right ( :to_the_right! ) |&gt; Z . right ( ) |&gt; Z . remove ( ) % Sourceror.Zipper { node : 3 , path : %{ parent : % Sourceror.Zipper { node : [ 2 , 3 ] , path : %{ parent : % Sourceror.Zipper { node : [ 1 , [ 2 , 3 ] , 4 , [ 5 , [ 6 , 7 ] ] , 8 ] , path : nil } , left : [ 1 ] , right : [ 4 , [ 5 , [ 6 , 7 ] ] , 8 ] } } , left : [ 2 ] , right : nil } } A couple interesting things happened. First, the item was removed and we can no longer see it in the zipper. Second, the focus moved to the right-most element inside the [2, 3] list. Whenever we remove an item, the focus is moved to the previous element in depth-first order. It's as if we did a Macro.postwalk step, but backwards. Let's look a what happens if we go higher in the tree: zipper |&gt; Z . down ( ) |&gt; Z . right ( ) |&gt; Z . insert_right ( :to_the_right! ) |&gt; Z . up ( ) % Sourceror.Zipper { node : [ 1 , [ 2 , 3 ] , :to_the_right! , 4 , [ 5 , [ 6 , 7 ] ] , 8 ] , path : nil } Instead of removing the added node, we went back up. The list was reconstructed and the :path is now nil , meaning we reached the top. One way is that when we go down a level, the tree is deconstructed to adjust to the new focus (the zipper is being opened), and when we go up a level the tree is reconstructed (the zipper is being closed). Now, while it's useful to know how to move up, down, left and right, this only works if we know the shape of the tree before hand. In the real world, we never know the shape of the tree, and so we use other kinds of movements to traverse it. For zippers, we have Z.next/1 and Z.prev/1 to move one step forward or backwards in depth-first pre-order, and Z.traverse/2-3 to perform a depth-first pre-order traversal while calling a function at each location. To demonstrate the ability of zippers to look around at the rest of the tree, we will traverse a tree of lists and integers, multiplying each integer by the integer at it's right, or removing it if it's already the rightmost sibling or the following element is another list: [ 1 , 2 , 3 , [ 4 , 5 , 6 , [ 7 , 8 , 9 ] , 10 , 11 ] , 12 , 13 , 14 ] |&gt; Z . zip ( ) |&gt; Z . traverse ( fn % Z { node : x } = zipper when is_integer ( x ) -&gt; with % Z { node : right } when is_integer ( right ) &lt;- Z . right ( zipper ) do Z . replace ( zipper , x * right ) else _ -&gt; Z . remove ( zipper ) end zipper -&gt; zipper end ) |&gt; Z . node ( ) [ 2 , 6 , [ 20 , 30 , ~c&quot;8H&quot; , 110 ] , 156 , 182 ] The traversal function takes a zipper representing the current position, and must return a new zipper. When we get back to the top of the tree, the second element will be the atom :end , signifying that the traversal ended, preventing subsequent calls to Z.next/1 from going down the tree again. The Z.node/1 function just extracts the current node from the zipper, so we can look at the final result instead of a zipper. That wraps up the introduction to zippers. We'll see them in action in the &quot;Multi alias expansion&quot; livebook, where we use them to simplify the traversal and transformations.","ref":"zippers.html#introduction","title":"Zippers - Introduction","type":"extras"},{"doc":"","ref":"expand_multi_alias.html","title":"Expand multi-alias syntax","type":"extras"},{"doc":"The multi-alias syntax alias Foo.{Bar, Baz} , while handy to alias multiple modules at once, makes module uses harder to search for in large code bases, as mentioned by credo warnings. We can use Sourceror to fix this issue by expanding this syntax into multiple calls to alias , each in its own line. Let's first start by installing Sourceror: Mix . install ( [ :sourceror ] ) And now lets parse an example to get an idea of the kind of structure we're going to work with: source = ~S&quot;&quot;&quot; defmodule Foo do alias Foo.{Bar, Baz} end &quot;&quot;&quot; quoted = Sourceror . parse_string! ( source ) We need to traverse the ast to find any occurrence of a qualified tuple call(ie. calls with the form {:., meta, [left, :{}]} as their first element) that is an argument to an :alias call. Then for each module inside of the curly brackets, we need to join the module segments from the left hand side with the ones in the right hand side, and finally put them in a call to :alias . For the traversal part, we can use Sourceror.postwalk/2 . Postwalk functions will go down the tree to the deepest node, then to the sibling nodes, then to the parent node, and so on until the whole tree is traversed. A way to think about it is that it traverses bottom to top, or that child nodes are always visited first. To convert a single alias into multiple ones, we need to extract the left side of the tuple and join it with the elements inside of the tuple. For the left part, we can extract it from the dot call we mentioned earlier. The first argument will always be the left hand side, and the second one the atom :{} . The elements inside the tuple are just the arguments of the qualified tuple call, ie the outer 3-tuple. Each of these elements will be an :__aliases__ call. In such calls, the arguments are the segments of the module as regular atoms, so for example the segments for Foo.Bar will be [:Foo, :Bar] . To create a module alias of the expanded Foo.{Bar} , we just need to join the segments and put them in an :__aliases__ call. Finally, that call needs to be wrapped in a call to :alias to effectively create an alias expression. Now we have a list of :alias calls, but the traversal function needs to return an AST node, not just a list(that would be considered a list literal). We can work around this for now by wrapping the aliases in a :__block__ and returning that. In other contexts like macros this would change the semantics of the code and it would not behave as we expect, but since we are doing these manipulations to output code as text, we can afford to do it: defmodule AliasExpansion do def expand_aliases ( quoted ) do Sourceror . postwalk ( quoted , fn { :alias , _ , [ { { :. , _ , [ left , :{} ] } , _ , right } ] } , state -&gt; { _ , _ , base } = left aliases = Enum . map ( right , fn { _ , _ , segments } -&gt; aliased = { :__aliases__ , [ ] , base ++ segments } { :alias , [ ] , [ aliased ] } end ) { { :__block__ , [ ] , aliases } , state } quoted , state -&gt; { quoted , state } end ) end end AliasExpansion . expand_aliases ( quoted ) |&gt; Sourceror . to_string ( ) |&gt; IO . puts ( ) This works for now because it was just a single expression in the whole module, but it will break as soon as we add more expressions: source = ~S&quot;&quot;&quot; defmodule Foo do alias Foo.{Bar, Baz} 42 end &quot;&quot;&quot; Sourceror . parse_string! ( source ) |&gt; AliasExpansion . expand_aliases ( ) |&gt; Sourceror . to_string ( ) |&gt; IO . puts ( ) Because we wrapped the aliases in a block, if we add more expressions the expanded aliases will be wrapped in parenthesis. This happens because do blocks contents are wrapped in a :__block__ whenever they have more than one expression, and what we are doing in our expansion is putting a block inside another block, so the formatter will interpret that as a block expression and will wrap it accordingly: Sourceror . parse_string! ( ~S&quot;&quot;&quot; def foo do :ok end &quot;&quot;&quot; ) |&gt; IO . inspect ( label : &quot;Single expression&quot; ) Sourceror . parse_string! ( ~S&quot;&quot;&quot; def foo do 42 :ok end &quot;&quot;&quot; ) |&gt; IO . inspect ( label : &quot;Multiple expressions&quot; ) One way to solve this issue is to mark our aliases block as a block that needs to be unwrapped if it's inside another block. When traversing, if we encounter a block, we reduce its arguments to unwrap any marked block, essentially &quot;adding multiple nodes&quot; to the block: defmodule AliasExpansionHandledMultiNodes do def expand_aliases ( quoted ) do Sourceror . postwalk ( quoted , fn { :alias , _ , [ { { :. , _ , [ _ , :{} ] } , _ , _ } ] } = quoted , state -&gt; aliases = expand_alias ( quoted ) { { :__block__ , [ unwrap_me? : true ] , aliases } , state } { :__block__ , meta , args } , state -&gt; args = Enum . reduce ( args , [ ] , &amp; unwrap_aliases / 2 ) { { :__block__ , meta , args } , state } quoted , state -&gt; { quoted , state } end ) end defp expand_alias ( { :alias , _ , [ { { :. , _ , [ left , :{} ] } , _ , right } ] } ) do { _ , _ , base } = left aliases = Enum . map ( right , fn { _ , _ , segments } -&gt; aliased = { :__aliases__ , [ ] , base ++ segments } { :alias , [ ] , [ aliased ] } end ) aliases end defp unwrap_aliases ( { :__block__ , [ unwrap_me? : true ] , aliases } , args ) do args ++ aliases end defp unwrap_aliases ( quoted , args ) do args ++ [ quoted ] end end Sourceror . parse_string! ( source ) |&gt; AliasExpansionHandledMultiNodes . expand_aliases ( ) |&gt; Sourceror . to_string ( ) |&gt; IO . puts ( )","ref":"expand_multi_alias.html#introduction","title":"Expand multi-alias syntax - Introduction","type":"extras"},{"doc":"Great! Now that we have addressed this issue, there is one last problem we need to take care of. In our current code we are completely ignoring the nodes metadata. This would be fine in most cases if we were working with macros, but it becomes a big issue if we want to turn this AST into formatted text. To avoid adding new types of AST nodes, Sourceror places comments in the nodes metadata, so if we discard nodes metadata, we would also be discarding it's associated comments. An important aspect of refactoring tools is that they should be able to preserve as much data as possible while doing the transformation, so we must take comments into account. This is easier to see with an example: ~S&quot;&quot;&quot; # Some comment alias Foo.{Bar, Baz} &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) |&gt; AliasExpansionHandledMultiNodes . expand_aliases ( ) |&gt; Sourceror . to_string ( ) |&gt; IO . puts ( ) This issue is easy to avoid if we always remember to pass the metadata around. In this particular example, due to the way Sourceror merges comments, we only need to preserve the alias and the individual :__aliases__ metadata. The other thing to keep is that we're getting rid of the first alias and starting anew with the right side segments. But this first alias is the one that holds the leading comments for the first expression, which means that if we discard it, we lose the comments right before the multi alias. We can solve this by attaching the leading comments to the first one, and the trailing comments to the last one: defmodule AliasExpansionHandledComments do def expand_aliases ( quoted ) do Sourceror . postwalk ( quoted , fn { :alias , _ , [ { { :. , _ , [ _ , :{} ] } , _ , _ } ] } = quoted , state -&gt; { aliases , state } = expand_alias ( quoted , state ) { { :__block__ , [ unwrap_me? : true ] , aliases } , state } { :__block__ , meta , args } , state -&gt; args = Enum . reduce ( args , [ ] , &amp; unwrap_aliases / 2 ) { { :__block__ , meta , args } , state } quoted , state -&gt; { quoted , state } end ) end defp unwrap_aliases ( { :__block__ , [ unwrap_me? : true ] , aliases } , args ) do args ++ aliases end defp unwrap_aliases ( quoted , args ) do args ++ [ quoted ] end defp expand_alias ( { :alias , alias_meta , [ { { :. , _ , [ left , :{} ] } , call_meta , right } ] } , state ) do { _ , _ , base_segments } = left leading_comments = alias_meta [ :leading_comments ] || [ ] trailing_comments = call_meta [ :trailing_comments ] || [ ] aliases = right |&gt; Enum . map ( &amp; segments_to_alias ( base_segments , &amp;1 ) ) |&gt; put_leading_comments ( leading_comments ) |&gt; put_trailing_comments ( trailing_comments ) { aliases , state } end defp segments_to_alias ( base_segments , { _ , meta , segments } ) do { :alias , meta , [ { :__aliases__ , [ ] , base_segments ++ segments } ] } end defp put_leading_comments ( [ first | rest ] , comments ) do [ Sourceror . prepend_comments ( first , comments ) | rest ] end defp put_trailing_comments ( list , comments ) do case List . pop_at ( list , - 1 ) do { nil , list } -&gt; list { last , list } -&gt; last = { :__block__ , [ trailing_comments : comments , # End of expression newlines higher than 1 will cause the formatter to add an # additional line break after the node. This is entirely optional and only showcased # here to improve the readability of the output end_of_expression : [ newlines : 2 ] ] , [ last ] } list ++ [ last ] end end end ~S&quot;&quot;&quot; # Some comment alias Foo.{Bar, Baz} &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) |&gt; AliasExpansionHandledComments . expand_aliases ( ) |&gt; Sourceror . to_string ( ) |&gt; IO . puts ( ) Older versions of sourceror required you to keep track of line numbers, and in an example like the one above you'd be required to keep track of how line numbers changed to Sourceror would apply corrections and ensure comments ended up in the correct places. Fortunately, newer Sourceror versions eliminate this issue so you can focus just on moving nodes around and stop worrying about hacky line numbers calculations. You can try with more convoluted examples and see that the code above produces the output you would expect: source = ~S&quot;&quot;&quot; defmodule Sample do # Some aliases alias Foo.{A, B, C, D, E, F} # Hello! alias Bar.{G, H, I, # Inner comment! # Inner comment 2! # Inner comment 3! J, # Comment for K! K # Comment for K 2! # Inner last comment! # Inner last comment 2! } # Not an inner comment def foo() do # Some scoped alias alias Baz.{A, B, C} # Just return :ok :ok # At the end end # Comment for :hello :hello end # End of file! &quot;&quot;&quot; source |&gt; Sourceror . parse_string! ( ) |&gt; AliasExpansionHandledComments . expand_aliases ( ) |&gt; Sourceror . to_string ( ) |&gt; IO . puts ( )","ref":"expand_multi_alias.html#handling-comments","title":"Expand multi-alias syntax - Handling comments","type":"extras"},{"doc":"So far the code works, but there is a detail that can be improved. When we expand the alias, we are placing all of them in temporary block with the :flatten_me? metadata flag, and then we remove it when going up in the postwalk. While this work, it is clearly a hack and we are working around one of the limitation of &quot;traditional&quot; tree traversals. Sourceror provides a Zipper implementation for the Elixir AST that makes it easy to traverse the AST in arbitrary directions and perform local manipulations like adding siblings or removing nodes. A tree Zipper is a data structure that represents a location in a tree, rather than the whole tree from top to bottom. It's shape is a 2-tuple, where the first element is the current node the zipper is focused on, and the second element is the context, or in other words, the rest of the tree from the perspective of the current node. When we are at the top level of alias Sourceror.Zipper , as : Z Z . zip ( [ 1 , [ 2 , 3 ] , 4 ] ) We get the tuple {[1, [2, 3], 4], nil} , where the first element is the tree and the second element is nil , signifying that this is a zipper focusing on the root node of the tree. To move around in the tree, we can use Z.up/1 and Z.down/1 for vertical movement, Z.left/1 and Z.right/1 for horizontal movement, and Z.next/1 and Z.prev/1 to walk the tree in a depth-first, pre-order fashion. Let's go down the tree and see what happens: zipper = Z . zip ( [ 1 , [ 2 , 3 ] , 4 ] ) zipper |&gt; Z . down ( ) Since we went down a level, we are now at the node 1 , as seen in the first element of the tuple, and the metadata is now a tree with the rest of the tree relative to 1 : l holds the sibling nodes at the left of the current node, r holds the sibling nodes at the right of the current node, and ptree holds a the parent tree as a zipper. In a way, every time we go down a level the zipper rips apart the parent zipper |&gt; Z . down ( ) |&gt; Z . up ( ) Most of the time we don't care about the metadata/context of the zipper, and we want to get the current node out of it. For that we can use Z.node/1 : zipper |&gt; Z . down ( ) |&gt; Z . node ( ) Or we can use Z.root/1 to go all the way up to zipper |&gt; Z . down ( ) |&gt; Z . root ( ) To move horizontally , Z.left/1 and Z.right/1 come to the rescue: zipper |&gt; Z . down ( ) |&gt; Z . right ( ) |&gt; Z . node ( ) zipper |&gt; Z . down ( ) |&gt; Z . right ( ) |&gt; Z . left ( ) |&gt; Z . node ( ) Finally, to walk the tree in a similar way to how we would walk it with Macro.prewalk , we can use Z.next/1 and Z.prev/1 : zipper |&gt; Z . next ( ) |&gt; Z . next ( ) |&gt; Z . next ( ) |&gt; Z . node ( ) zipper |&gt; Z . next ( ) |&gt; Z . next ( ) |&gt; Z . next ( ) |&gt; Z . prev ( ) |&gt; Z . prev ( ) |&gt; Z . node ( ) Walking a tree is nice, but we also need to change it. For that we have tools like Z.replace/2 , Z.update/2 , Z.remove/1 , Z.insert_child or Z.traverse/3 . We can see them in action by rewriting our AliasExpansion module to use the zipper api: defmodule AliasExpansionWithZipper do alias Sourceror.Zipper , as : Z def expand_aliases ( quoted ) do Z . zip ( quoted ) |&gt; Z . traverse ( &amp; expand_alias / 1 ) |&gt; Z . root ( ) end defp expand_alias ( % Z { node : { :alias , alias_meta , [ { { :. , _ , [ left , :{} ] } , call_meta , right } ] } } = zipper ) do { _ , _ , base_segments } = left leading_comments = alias_meta [ :leading_comments ] || [ ] trailing_comments = call_meta [ :trailing_comments ] || [ ] right |&gt; Enum . map ( &amp; segments_to_alias ( base_segments , &amp;1 ) ) |&gt; put_leading_comments ( leading_comments ) |&gt; put_trailing_comments ( trailing_comments ) |&gt; Enum . reverse ( ) |&gt; Enum . reduce ( zipper , &amp; Z . insert_right ( &amp;2 , &amp;1 ) ) |&gt; Z . remove ( ) end defp expand_alias ( zipper ) , do : zipper defp segments_to_alias ( base_segments , { _ , meta , segments } ) do { :alias , meta , [ { :__aliases__ , [ ] , base_segments ++ segments } ] } end defp put_leading_comments ( [ first | rest ] , comments ) do [ Sourceror . prepend_comments ( first , comments ) | rest ] end defp put_trailing_comments ( list , comments ) do case List . pop_at ( list , - 1 ) do { nil , list } -&gt; list { last , list } -&gt; last = { :__block__ , [ trailing_comments : comments , end_of_expression : [ newlines : 2 ] ] , [ last ] } list ++ [ last ] end end end source |&gt; Sourceror . parse_string! ( ) |&gt; AliasExpansionWithZipper . expand_aliases ( ) |&gt; Sourceror . to_string ( ) |&gt; IO . puts ( ) The output is the same. Let's confirm: without_zipper = source |&gt; Sourceror . parse_string! ( ) |&gt; AliasExpansionHandledComments . expand_aliases ( ) |&gt; Sourceror . to_string ( ) with_zipper = source |&gt; Sourceror . parse_string! ( ) |&gt; AliasExpansionWithZipper . expand_aliases ( ) |&gt; Sourceror . to_string ( ) without_zipper == with_zipper The usage of a zipper greatly simplifies the code. We no longer have to resort to hacks like grouping the aliases in a block with the :flatten_me? metadata and wait for the traversal to go up to unwrap it. We just say &quot;add these nodes after me&quot; and that's it. The only detail is that we need to reverse the aliases so the reducing step inserts them in the correct order, but otherwise the code is much more straightforward: find an alias, generate the expanded aliases, add them to the tree and remove the old alias.","ref":"expand_multi_alias.html#simplifying-traversals-with-sourceror-zipper","title":"Expand multi-alias syntax - Simplifying traversals with Sourceror.Zipper","type":"extras"},{"doc":"We've seen how Sourceror approaches comments, we looked at how to perform a simple but yet non-trivial transformation to the source code, and finally we had a look into the Zipper API to make powerful transformations in a simple and straightforward way. With these primites, you are now able to leverage Sourceror to create your own transformations, refactoring scripts and maybe even macros.","ref":"expand_multi_alias.html#final-words","title":"Expand multi-alias syntax - Final words","type":"extras"}]